open Sys
open Prank
open Util

let usage_text = "usage prank_learn [options] <train_file>
[options] are any of: 
   -r <rounds>                         number of training rounds (default 5)
   -v                                  will use voting in decoding.
   -W <weights_file>                   prints a human-readable weights file
   -#feats <num_features>              otherwise will infer from data
   -#ranks <num_ranks>                   \"\"                  \"\"   
   -test <test_file> <output_file>     for immediate testing.
   -model <model_file>                 write model

one of -test or -model must be specified (otherwise, what's the point??)
"

let arg_error str = 
  Printf.fprintf stderr "%s\n" usage_text;
  failwith ("Error: " ^ str)

let l = Array.length argv
let _ = if l < 2 then arg_error "not enough arguments"
let rounds = ref 5
let vote = ref false
let avg = ref false
let norm = ref false
let norm2 = ref false
let weights_file = ref None
let num_ranks = ref None
let num_feats = ref None
let train_file = open_in argv.(l-1)
let model_file = ref None
let test_file = ref None
let out_file = ref None

let _ =
  let rec read_opt p = 
    if p >= l-1 then () else
      match argv.(p) with
	| "-r" -> 
	    ((try rounds := int_of_string argv.(p+1)  
	      with _ -> arg_error "-r expects int arg") 
	    ;read_opt(p+2))
	| "-W" -> 
	    ((try weights_file := Some (open_out argv.(p+1)) 
	      with _ -> arg_error "-W expects file arg")
	    ;read_opt(p+2))
	| "-#feats" -> 
	    ((try num_feats := Some (int_of_string argv.(p+1)) 
	      with _ -> arg_error "-#feats expects int arg")
	    ;read_opt(p+2))
	| "-#ranks" ->
	    ((try num_ranks := Some (int_of_string argv.(p+1)) 
	      with _ -> arg_error "-#ranks expects int arg")
	    ;read_opt(p+2))
	| "-model" ->
	    ((try model_file := Some (open_out argv.(p+1)) 
	      with _ -> arg_error "-model expects file arg")
	    ;read_opt(p+2))
	| "-test" ->
	    ((try test_file := Some (open_in argv.(p+1)) 
	      with _ -> arg_error "-test expects 2 file args")
	    ;(try out_file := Some (open_out argv.(p+2)) 
	      with _ -> arg_error "-test expects 2 file args")
	    ; read_opt(p+3))
	| "-v" -> 
	    (vote := true 
	    ;read_opt(p+1))
	| "-a" -> 
	    (avg := true 
	    ;read_opt(p+1))
	| _ -> arg_error ("unknown option: '" ^ argv.(p) ^ "'") in
  read_opt 1 ;;

match !test_file, !model_file with
  | None, None -> arg_error 
      "one of -test or -model must be specified (otherwise, what's the point??)"
  | _ -> () 

let data = fst (parse_data_file train_file)

let num_feats' = match !num_feats with 
  | None -> get_highest_feature data + 1
  | Some n -> n 

let num_ranks' = match !num_ranks with
  | None -> List.fold_left 
      (fun prev instance -> if instance.label > prev then instance.label else prev) (-5) data
  | Some n -> n 

let model = 
  { w = Array.make num_feats' 0. ;
    b = Array.make (num_ranks' - 1) 0. ;
    w_avg = Array.make num_feats' 0. ;
    b_avg = Array.make (num_ranks' - 1) 0. ;
    w_sum = Array.make num_feats' 0. ;
    b_sum = Array.make (num_ranks' - 1) 0. ;
    w_hist = Dray.make 10000 ;
    b_hist = Dray.make 10000 ;
    hist_count = Dray.make 10000;
    vote = !vote ;
    avg = !avg;
    normalize = !norm;
    normalize2 = !norm2;
    trained = false ;
    updates = ref 0;
    num_ranks = num_ranks' ;
    num_feats = num_feats' }

(* updates the model with the weighted instance features *)
let update_with_instance (model : single_model) (features : features) : unit =
  let w, vote = model.w, model.vote in
  let f feat feat_val = w.(feat) <- w.(feat) +. feat_val in
  Hashtbl.iter f features ;
  if vote then
    Dray.add model.w_hist features

let update_boundaries (model : single_model) (instance : instance) prediction =
  let label, b = instance.label, model.b in
  if prediction < label then
    for i=prediction-1 to label-2 do
      b.(i) <- b.(i) -. 1.
    done
  else
    for i=label-1 to prediction-2 do
      b.(i) <- b.(i) +. 1.
    done ;
  if model.vote then
    let b_copy = Array.init (model.num_ranks - 1) (fun i -> b.(i)) in
    Dray.add model.b_hist b_copy
      
let train_with_instance (model : single_model) (mistakes : float ref) (instance : instance) =
  let prediction = predict instance model in
  if prediction != instance.label then
    ( mistakes := !mistakes +. 1.;
      let instance' = Hashtbl.create 15 in
      let penalty = float_of_int (instance.label - prediction) in
      Hashtbl.iter 
	(fun feat feat_val -> Hashtbl.add instance' feat (feat_val *. penalty)) 
	instance.features ;
      update_with_instance model instance' ;
      update_boundaries model instance prediction )

let finalize_model (model : single_model) =
  model.trained <- true 

let definalize_model (model : single_model) = 
  model.trained <- false

let _ = 
  let rec learn_model (data : data) (model : single_model) (round : int) =
    if round > !rounds then 
      (finalize_model model;
       prerr_newline ()) 
    else
      let mistakes = ref 0. in
      prerr_endline (Printf.sprintf 
			"\n=============== ITERATION %d =================" round) ;
      List.iter (train_with_instance model mistakes) data ; 
      prerr_endline (Printf.sprintf "\ntrain accuracy: %f"  
			(1. -. (!mistakes /. (float_of_int (List.length data))))) ;
      learn_model data model (round+1)
  in
  learn_model data model 1 ;
  (match !model_file with 
    | Some out_ch -> Marshal.to_channel out_ch (Single model) [] ;
    | _ -> ()) ;
  (match !test_file, !out_file with 
    | Some test_ch, Some out_ch -> run_model (Single model) (fst (parse_data_file test_ch)) out_ch
    | _ -> ()) ;
  (match !weights_file with
    | Some ch -> 
	let write_w w_val = Printf.fprintf ch "%f\n" w_val in
	Array.iter write_w model.w
    | _ -> ())


	
	
	
    
	    
