(*pp $PP *)
open Sys
open Prank
open Util
open Printf

module Dray = DynArray
module Opt = Option
module H = ExtHashtbl.Hashtbl

let debug = Prank.debug
let rounds = ref 5
let curve = ref max_int
let vote = ref false
let avg = ref false
let norm = ref false
let norm2 = ref false
let share_boundaries = ref false
let use_agree = ref false
let each = ref false
let shuffle = ref false
let agree_file = ref None
let weights_file = ref None
let train_file = ref None
let model_file = ref None
let test_file = ref None
let out_file = ref (Some (open_out "multi.output"))
let out_null = open_out "/dev/null"
let con_train = ref None  
let con_test = ref None
let agree_train = (ref None : (int list * float) list list option ref) 
let agree_test = ref None
let m_train = ref None
let m_test = ref None
let m_type = ref None
let seed = ref None
let alpha = ref 0.5
let beta = H.create 20
let sim = H.create 20
let use_sim = ref false
let meta_dist = ref 0
let food_atmos = ref false

let set_seed s = seed := Some s
let set_out r s = r := Some (open_out s)
let set_in r s = r := Some (open_in s)
let set_concensus_scores fname = 
  let l, tag = L.map S.strip (list_of_file fname), ref "" in
  let l' = L.map (fun ln -> if ln.[0] = '#' then tag := ln ; (ln,!tag)) l in
  let train = L.filter (fun (ln,tag) -> tag = "#TRAIN" && ln.[0] <> '#') l'
  and test = L.filter (fun (ln,tag) -> tag = "#TEST" && ln.[0] <> '#') l' in
  if not (
      L.length train + L.length test = L.length l' - 2 ||
      L.length train + L.length test = L.length l' - 1
    ) then failwith "#TRAIN and #TEST tags not set properly in concensus data";
  con_train := Some (L.map (fun (ln,_) -> float_of_string ln) train);
  con_test := Some (L.map (fun (ln,_) -> float_of_string ln) test)

let set_meta_scores fname =
  let l, tag = L.map S.strip (list_of_file fname), ref "" in
  let l' = L.map (fun ln -> if ln.[0] = '#' then tag := ln ; (ln,!tag)) l in
  let train = L.filter (fun (ln,tag) -> tag = "#TRAIN" && ln.[0] <> '#') l'
  and test = L.filter (fun (ln,tag) -> tag = "#TEST" && ln.[0] <> '#') l' in
  if not (
      L.length train + L.length test = L.length l' - 2 ||
      L.length train + L.length test = L.length l' - 1
    ) then failwith "#TRAIN and #TEST tags not set properly in concensus data";
  m_train := Some (L.map (fun (ln,_) -> A.of_list (L.map float_of_string (split ln))) train);
  m_test := Some (L.map (fun (ln,_) -> A.of_list (L.map float_of_string (split ln))) test)

let set_agree_scores fname =
  let err = "#TRAIN and #TEST tags not set properly in agreement data" in
  let l, tag = L.map S.strip (list_of_file fname), ref "" in
  let l' = L.map (fun ln -> if ln.[0] = '#' then tag := ln ; (ln,!tag)) l in
  let train = L.filter (fun (ln,tag) -> S.starts_with tag "#TRAIN" && ln.[0] <> '#') l'
  and test = L.filter (fun (ln,tag) -> S.starts_with tag "#TEST" && ln.[0] <> '#') l' in
  let h_train,h_test = H.create 5000, H.create 5000 in
  let f h (ln,tag) = match split tag with
    | _ :: tl -> (try
	  let key, f = L.sort (L.map int_of_string tl), float_of_string ln in
	  H.replace h key (f :: (entry_get h key [])) 
      with _ -> failwith (err ^ " (2)"))
    | _ -> failwith (err ^ " (3)") in
  L.iter (f h_train) train;
  L.iter (f h_test) test;
  let f h =
    let keys = A.of_enum (H.keys h) in
    let d = A.map (fun k -> (A.of_list (L.rev (H.find h k)))) keys in
    let res = Dray.make 5000 in
    for i = 0 to A.length d.(0) - 1 do
      let e = A.init 
	(A.length d) 
	(fun j -> (keys.(j), d.(j).(i))) in
      Dray.add res (A.to_list e) 
    done ;
    Dray.to_list res in
  let tmp1 = f h_train
  and tmp2 = f h_test in
  agree_train := Some tmp1;
  agree_test := Some tmp2
    
let set_beta fname = 
  let err = "Wrong file format for 'beta' file." in
  let l = L.map S.strip (list_of_file fname) in
  L.iter 
    (fun ln -> if ln.[0] <> '#' then 
	match split ln with
	  | [k;f] -> let key = (L.sort (L.map int_of_string (Str.split (Str.regexp "_") k))) in
		     H.replace beta key (float_of_string f)
	  | _ -> failwith err)
    l
    
let set_sim fname =
  use_sim := true;
  let err = "Wrong file format for 'sim' file." in
  let l = L.map S.strip (list_of_file fname) in
  L.iter
    (fun ln ->
      match split ln with
	| [a';b';c'] -> 
	    (try
		let a,b,c = int_of_string a', int_of_string b', float_of_string c' in
		H.replace sim (a,b) c;
		H.replace sim (b,a) c;
	      with _ -> failwith err)
	| _ -> failwith err)
    l
		
      
let usage_msg = "USAGE:\n\nmultiprank_learn [options] train"
let speclist = Arg.align
  [ ("-r", Arg.Set_int rounds, "<rounds> Number of training rounds (default 5)");
    ("-W", Arg.String (set_out weights_file), "<file> Write human-readable weights file");
    ("-model", Arg.String (set_out model_file), "<file> Write model");
    ("-test", Arg.String (set_in test_file), "<file> Test file");
    ("-test.out", Arg.String (set_out out_file), "<file> Write results (default 'multi.out')");
    ("-each", Arg.Set each, " Run model on test data after each round");
    ("-v", Arg.Set vote, " Use voting during decoding");
    ("-a", Arg.Set avg, " Use averaging during decoding");
    ("-norm", Arg.Set norm, " Normalize averaged weight vectors by number of updates");
    ("-norm2", Arg.Set norm2, " Normalize averaged weight vectors by norm");
    ("-b", Arg.Set share_boundaries, " Share boundaries");
    ("-agree", Arg.Set use_agree, " Train and use new agreement model");
    ("-agree.model", Arg.String (set_in agree_file),"<file> Use perceptron agreement model");
    ("-concensus", Arg.String set_concensus_scores,"<file> Use precomputed consensus scores");
    ("-agree.data", Arg.String set_agree_scores,"<file> Use precomputed agree scores");
    ("-meta", Arg.String set_meta_scores, "<file> Use precomputed meta-model scores");
    ("-dist", Arg.Set_int meta_dist, "<int> Distance Meta type");
    ("-food_atmos", Arg.Set food_atmos, " Use food-atmosphere divergence meta model");
    ("-alpha", Arg.Set_float alpha, "<float> Multiplier of concensus model (default 0.5)");
    ("-curve", Arg.Set_int curve, "<int> Number of training instances to use (default all)");
    ("-beta", Arg.String set_beta,"<file> File specifying multipliers of agreement model");
    ("-s", Arg.Set shuffle, " Shuffle training data");
    ("-seed", Arg.Int set_seed, "<int> Seed for shuffling training data");
    ("-sim", Arg.String set_sim,"<file> File specifying task similarities to use")
  ];;

let arg_error str = 
  Arg.usage speclist usage_msg;
  prerr_newline ();
  failwith ("Error: " ^ str) ;;

Arg.parse speclist (set_in train_file) usage_msg;;

if Opt.is_none !test_file && Opt.is_none !model_file then
  arg_error "One of -test or -model must be specified (otherwise, what's the point??)";;

if Opt.is_none !train_file then 
  arg_error "Training file must be set!";;

if !meta_dist <> 0 then
  m_type := Some (Dist !meta_dist);;

if !food_atmos then
  m_type := Some FoodAtmosDiv;;

if (!m_train <> None && !m_type = None) || (!m_train = None && !m_type <> None) then
  arg_error "Meta-type has to be specified along with meta scores";;

if !agree_train <> None then
  agree_train := Some (L.map 
			  (fun l -> L.map (fun (t,f) -> (t, f *. !alpha *. entry_get beta t 1.)) l) 
			  (Opt.get !agree_train));;

if !agree_test <> None then
  agree_test := Some (L.map 
			 (fun l -> L.map (fun (t,f) -> (t, f *. !alpha *. entry_get beta t 1.)) l) 
			 (Opt.get !agree_test));;

if !m_train <> None then
  m_train := Some (L.map (A.map (fun f -> f *. !alpha)) (Opt.get !m_train));;

if !m_test <> None then
  m_test := Some (L.map (A.map (fun f -> f *. !alpha)) (Opt.get !m_test));

if !con_train <> None then
  con_train := Some (L.map (fun f -> f *. !alpha) (Opt.get !con_train));;

if !con_test <> None then
  con_test := Some (L.map (fun f -> f *. !alpha) (Opt.get !con_test));;

let (data', num_categories) = parse_data_file ~multi:true (Opt.get !train_file)
let data' = if !shuffle || !seed <> None 
  then shuffle_list ~seed:!seed data' 
  else data';;

if !curve != max_int then
  curve := !curve * num_categories;;
let data = if !curve >= L.length data' then data' else 
    A.to_list (A.sub (A.of_list data') 0 !curve)
let num_feats = get_highest_feature data + 1
let num_ranks = L.fold_left 
  (fun prev instance -> if instance.label > prev then instance.label else prev) (-5) data

let make_boundaries () = (* A.make (num_ranks - 1) 0. *)
  let half = (num_ranks - 1) / 2 in
  A.init (num_ranks - 1) (fun i -> float (i - half))

let init_model boundaries = 
  let m =
    { w = A.make num_feats 0. ;
      b = if !share_boundaries then boundaries else make_boundaries () ;
      w_hist = (let z = Dray.make 10000 in Dray.add z empty_instance.features; z);
      b_hist = (let z = Dray.make 10000 in Dray.add z (A.make (num_ranks - 1) 0.); z);
      hist_count = (let z = Dray.make 10000 in Dray.add z (ref 1); z);
      w_sum = A.make num_feats 0.;
      b_sum = A.make (num_ranks - 1) 0.;
      w_avg = A.make num_feats 0.;
      b_avg = A.make (num_ranks - 1) 0.;
      vote = !vote;
      avg = !avg;
      normalize = !norm;
      normalize2 = !norm2;
      trained = false;
      updates = ref 1;
      num_ranks = num_ranks;
      num_feats = num_feats } in
  if not !norm then (m.w_avg <- m.w_sum; m.b_avg <- m.b_sum);
  m
			    
			    

let boundaries = make_boundaries ()
let a_model = Opt.map Marshal.from_channel !agree_file

let model = 
  { models = (A.init num_categories (fun _ -> init_model boundaries));
    num_categories = num_categories;
    multi_trained = false;
    agree = if Opt.is_none a_model && !use_agree then
	Some { Perceptron.weights = A.make num_feats 0. ;
	       avg_weights = A.make num_feats 0. ;
	       bias = None;
	       avg_bias = None;	    
	       kernel = None ;
	       alpha = H.create 1 ;
	       avg = true ;
	       history =  [(empty_instance,0)] ;
	       trained = false }
      else a_model;
    c_train = !con_train;
    c_test = !con_test;
    a_train = !agree_train;
    a_test = !agree_test;
    m_train = !m_train;
    m_test = !m_test;
    m_type = !m_type;
    alpha = !alpha;
    num_updates = ref 0;
    use_sim = !use_sim;
    sim = sim
  }

(* updates the model with the weighted instance features *)
let update_with_instance (model : multi_model) (features : features) category : unit =
  let s_model = model.models.(category) in
  incr s_model.updates;
  let w, vote, avg, w_sum = s_model.w, s_model.vote, s_model.avg, s_model.w_sum in
  let f feat feat_val = w.(feat) <- w.(feat) +. feat_val in
  H.iter f features ;
  if avg then 
    A.iteri (fun i v -> w_sum.(i) <- w_sum.(i) +. v) w
  else if vote then
    begin
      Dray.add s_model.w_hist features;
      Dray.add s_model.hist_count (ref 0);
      A.iter 
	(fun m -> incr (Dray.last m.hist_count))
	model.models
    end

let update_boundaries (model : multi_model) (instance : instance) prediction category =
  let label, b, s_model = instance.label, model.models.(category).b, model.models.(category) in
  if prediction < label then
    for i=prediction-1 to label-2 do
      b.(i) <- b.(i) -. 1.
    done
  else
    for i=label-1 to prediction-2 do
      b.(i) <- b.(i) +. 1.
    done ;
  if s_model.avg then
    A.iteri (fun i v -> s_model.b_sum.(i) <- s_model.b_sum.(i) +. v) b
  else if s_model.vote then
    let b_copy = A.init (s_model.num_ranks - 1) (fun i -> b.(i)) in
    Dray.add s_model.b_hist b_copy

      
let train_with_prediction  
    (model : multi_model) (mistakes : float ref) instance category prediction =
  if not model.multi_trained && prediction <> instance.label then
    begin
      mistakes := !mistakes +. 1.;
      let instance' = H.create 15 
      and penalty = float_of_int (instance.label - prediction) in
      H.iter 
	(fun feat feat_val -> H.add instance' feat (feat_val *. penalty)) 
	instance.features ;
      update_with_instance model instance' category;
      update_boundaries model instance prediction category ;
      incr model.num_updates;
    end

let train_with_instances ?(c_score=None) ?(a_scores=None) ?(m_scores=None) 
    model mistakes instances =
  let predictions = predict_multi ~c_score ~a_scores ~m_scores instances model 
  and labels = A.map (fun i->i.label) instances in
  if debug then 
    print_endline (sprintf "~~: %d%d%d%d%d" predictions.(0) predictions.(1) 
		      predictions.(2) predictions.(3) predictions.(4)); 
  A.iteri
    (fun i p -> train_with_prediction model mistakes instances.(i) i p)
    predictions;
  match model.agree with
    | Some a_model ->
	if not a_model.Perceptron.trained then
	  let int_of_bool b = if b then 1 else -1 in
	  let same_p, same_l = all_same predictions, all_same labels in
	  let a_instance = { label = int_of_bool same_l; 
			       features = instances.(0).features;
			       info = instances.(0).info } in
	  let a_prediction = int_of_bool same_p in 
	  Perceptron.update_with_prediction a_model a_instance a_prediction
    | _ -> ()
	  

let finalize_model (model : multi_model) =
  model.multi_trained <- true;
  A.iter 
    (fun m -> m.trained <- true; 
      if m.normalize then normalize m; 
      if m.normalize2 then normalize2 m)
    model.models; 
  match model.agree with
    | Some a_model -> Perceptron.finalize_model a_model
    | None -> ()  
	
let definalize_model (model : multi_model) =
  model.multi_trained <- false;
  A.iter (fun m -> m.trained <- false) model.models; 
  match model.agree with
    | Some a_model -> Perceptron.definalize_model a_model
    | None -> ()  

  
let _ = 
  let rec learn_models (data : data) (model : multi_model) (round : int) =
    let mistakes, instances, i = ref 0., A.make model.num_categories empty_instance, ref 0  in
    print_endline (Printf.sprintf "\n=============== ITERATION %d =================" round) ;
    let c_score_enum = Opt.map_default L.enum (Enum.empty ()) model.c_train 
    and a_score_enum = Opt.map_default L.enum (Enum.empty ()) model.a_train 
    and m_score_enum = Opt.map_default L.enum (Enum.empty ()) model.m_train in
    L.iter 
      (fun inst -> 
	instances.(!i) <- inst; 
	incr i; 
	if !i = model.num_categories then 
	  (i := 0; 
	   train_with_instances model 
	     ~a_scores:(Enum.get a_score_enum) ~c_score:(Enum.get c_score_enum) 
	     ~m_scores:(Enum.get m_score_enum) mistakes instances))
      data;
    print_endline 
      (Printf.sprintf "\ntrain err: %f" (!mistakes /. (float_of_int (L.length data))));
    if !each then
      (match !test_file with
	| Some test_ch ->
	    seek_in test_ch 0;
	    finalize_model model;
	    run_model (Multi model) (fst (parse_data_file ~multi:true test_ch)) out_null;
	    definalize_model model;
	| _ -> ());
    if model.models.(0).normalize2 then A.iter normalize3 model.models;
    if !mistakes > 0. && round < !rounds then
      learn_models data model (round+1)
    else 
      begin
	finalize_model model;
	print_newline ()
      end
  in
  learn_models data model 1 ;
  aaa := 0; ada := 0; add := 0; ddd := 0; daa := 0; dad := 0;
  (match !model_file with 
    | Some out_ch -> Marshal.to_channel out_ch (Multi model) [] ;
    | _ -> ()) ;
  (match !test_file, !out_file with 
    | Some test_ch, Some out_ch -> 
	seek_in test_ch 0;
	run_model ~verbose:0 (Multi model) (fst (parse_data_file ~multi:true test_ch)) out_ch
    | _ -> ()) ;
  (match !weights_file with
    | Some ch -> 
	let write_w w_val = Printf.fprintf ch "%f\n" w_val in
	A.iter write_w model.models.(0).w  (* XXX *) 
    | _ -> ()) ;
  Printf.printf "\naaa  %d\nada  %d\nadd  %d\nddd  %d\ndaa  %d\ndad  %d\n" 
    !aaa !ada !add !ddd !daa !dad  
    
