open Util

type alpha = (instance,int) Hashtbl.t
type weights = float array
type kernel = int option
type history = (instance * int) list
type model = 
    { weights : weights;
      mutable bias : float option;
      avg_weights : weights;
      mutable avg_bias : float option;
      kernel : kernel;
      alpha : alpha ;
      mutable history : history;
      avg : bool ;
      mutable trained : bool }


(*********************************************** DECODING **)

(* polynomial kernel function *)
let poly_kernel (features1 : features) (features2 : features) (d : int) : float = 
  let dp = dot_prd_ff features1 features2 in
  (1. +. dp)**(float_of_int d)
      
let get_score_straight (instance : instance) (model : model) = 
  let weights = if (model.avg && model.trained) then model.avg_weights else model.weights in
  dot_prd_fw instance.features weights

let get_score_poly_kernel (instance : instance) (model : model) (d : int) =
  let alpha = model.alpha in
  let f train_instance alpha_val sum = 
    let score = poly_kernel instance.features train_instance.features d in
    (float_of_int alpha_val) *. score +. sum in
  Hashtbl.fold f alpha 0.

let get_vote_poly_kernel (instance : instance) (model : model) (d : int) = 
  let pos = ref 0 in
  let neg = ref 0 in
  let f prev_score (train_instance,surv) =
    let score = poly_kernel instance.features train_instance.features d in
    let new_score = prev_score +. (float_of_int train_instance.label)*.score in
    if new_score > 0. then
      pos := !pos + surv
    else
      neg := !neg + surv ;
    new_score
  in
  let _ = List.fold_left f 0. model.history in
  if !pos > !neg then 1. else -1.
    
let get_score instance model = 
  match model.kernel with
      None   -> get_score_straight instance model
    | Some d -> 
	if (model.avg && model.trained) then 
	  get_vote_poly_kernel instance model d
	else
	  get_score_poly_kernel instance model d
	  
let predict instance model : label = 
  let score = get_score instance model
  and bias = if (model.avg && model.trained) then 
      Option.default 0. model.avg_bias
    else 
      Option.default 0. model.bias in
  if score > bias then 1 else -1

(*****************************************************  LEARNING **)

(* updates model weights in place *)
let update_weights (model : model) (instance : instance) : unit =
  let weights, avg_weights = model.weights, model.avg_weights in
  let f feat feat_val = 
    weights.(feat) <- weights.(feat) +. (float instance.label) *. feat_val in
  Hashtbl.iter f instance.features ;
  if model.avg && Opt.is_none model.kernel then
    let g i w_val = avg_weights.(i) <- avg_weights.(i) +. w_val in
    Array.iteri g weights
      
(* updates model alpha in place *)
let update_alpha (model : model) (instance : instance) : unit =
  entry_add model.alpha instance instance.label

let update_with_prediction ?(mistakes=ref 0.) model instance prediction =
  if prediction = instance.label then
    begin
      if model.avg && Opt.is_some model.kernel then
	let hd = List.hd model.history in
	let tl = List.tl model.history in
	model.history <- (fst hd, snd hd + 1) :: tl (* survived once more.. *)
    end
  else
    begin
      mistakes := !mistakes +. 1.;
      model.bias <- Option.map ((+.) (float prediction)) model.bias;
      model.avg_bias <- Option.map ((+.) (Option.default 0. model.bias)) model.avg_bias;
      match model.kernel with
	| None -> update_weights model instance         (* don't care bout hist *)
	| Some _ -> (update_alpha model instance ;
		     model.history <- (instance,1) :: model.history ) (* add new instance to history *)
    end
  
let update mistakes model instance : unit =
(*   print_int (int_of_float !mistakes); print_string "..."; *)
  if not model.trained then
    let prediction = predict instance model in
    update_with_prediction ~mistakes model instance prediction
    
let finalize_model model =
  model.history <- (List.rev model.history) ;
  model.trained <- true

let definalize_model model =
  model.history <- (List.rev model.history) ;
  model.trained <- false 


  
  
  
    
    
    

   
  
