(*pp $PP *)
open Util
open Printf

module Dray = DynArray
module Opt = Option
module A = ExtArray.Array
module L = ExtList.List
 
let debug = false
let counter = ref 0
let aaa = ref 0 (* temporary global state tracker!!! *)
let ada = ref 0
let add = ref 0
let ddd = ref 0
let daa = ref 0
let dad = ref 0 

let all_ranks_glob = Global.empty "list of all possible rank arrays"

type boundaries = float array
type weights = float array
type single_model = 
    { w : weights; 
      b : boundaries;
      w_sum : weights;
      b_sum : boundaries;
      mutable w_avg : weights;
      mutable b_avg : boundaries;
      w_hist : features Dray.t;
      b_hist : boundaries Dray.t;
      hist_count : int ref Dray.t;
      vote : bool;
      avg : bool;
      normalize : bool;
      normalize2 : bool;
      mutable trained : bool;       (* whether or not finished training yet *)
      updates : int ref;
      num_ranks : int;
      num_feats : int }

type meta_model = Dist of int | FoodAtmosDiv
type agreement_model = Perceptron.model
type multi_model =
    { models : single_model array;
      mutable multi_trained : bool;
      agree : agreement_model option; 
      num_categories : int;
      num_updates : int ref;
      c_train : float list option;     (*concensus data*)
      c_test : float list option; 
      a_train : ((int list) * float) list list option;
      a_test : ((int list) * float) list list option;
      m_train : float array list option;
      m_test : float array list option;
      m_type : meta_model option;
      alpha : float;
      use_sim : bool;
      sim : ((int * int),float) H.t;
    }

type model = Multi of multi_model | Single of single_model

let multi_of_single (model : single_model) = 
  { models = [|model|] ;
    multi_trained = true ;
    num_categories = 1 ;
    agree = None ;
    c_train = None ;
    c_test = None;
    a_train = None;
    a_test = None ;
    m_train = None;
    m_test = None;
    m_type = None;
    alpha = 1.0 ;
    num_updates = ref 0;
    use_sim = false;
    sim = H.create 0
  }

(************************************************************ DECODING **)

let normalize (m : single_model) =
  let updates = float(!(m.updates)) in
  let b_sum, w_sum, b_avg, w_avg = m.b_sum, m.w_sum, m.b_avg, m.w_avg in
  A.iteri 
    (fun i v -> w_avg.(i) <- v /. updates)
    w_sum; 
  A.iteri 
    (fun i v -> b_avg.(i) <- v /. updates)
    b_sum
  
let normalize2 (m : single_model) = 
  let c = 100. in
  let b_sum, w_sum, b_avg, w_avg = m.b_sum, m.w_sum, m.b_avg, m.w_avg 
  and n = ref 0. in
  A.iter (fun v -> n := !n +. v*.v) w_sum;
  n := (sqrt !n) /. c;
  A.iteri
    (fun i v -> w_avg.(i) <- v /. !n)
    w_sum;
  A.iteri
    (fun i v -> b_avg.(i) <- v /. !n)
    b_sum
      
let normalize3 (m : single_model) =
  let c = 100. in
  let w, b = m.w, m.b 
  and n = ref 0. in
  A.iter (fun v -> n := !n +. v*.v) w;
  n := (sqrt !n) /. c;
  A.iteri (fun i v -> w.(i) <- v /. !n) w;
  A.iteri (fun i v -> b.(i) <- v /. !n) b
  
let get_rank (score : float) (boundaries : boundaries) =
  let res, i = ref (-1), ref (A.length boundaries - 1) in
  while !res = -1 && !i >= 0 do
    if score > boundaries.(!i) then
      res := (!i + 2)
    else
      decr i
  done ;
  abs(!res) 


let predict instance (model : single_model) =
  if (not model.vote && not model.avg) || (not model.trained) then
    let score = dot_prd_fw instance.features model.w in
    get_rank score model.b
  else if model.avg then
    let score = dot_prd_fw instance.features model.w_avg in
    get_rank score model.b_avg
  else
    begin
      let votes, features, w_hist, b_hist, sum = 
	(A.make model.num_ranks 0), instance.features, 
	model.w_hist, model.b_hist, ref 0. in
      let add_vote (w : features) (b : boundaries) : unit = 
	let score = (dot_prd_ff features w) +. !sum in
	let rank = get_rank score b in
	votes.(rank - 1) <- votes.(rank - 1) + 1 ;
	sum := score in
      Enum.iter2 add_vote (Dray.enum w_hist) (Dray.enum b_hist) ;
      let best = (ref 0, ref votes.(0)) in
      for i = 1 to A.length votes - 1 do
	if votes.(i) >= !(snd best) then
	  (fst best := i; snd best := votes.(i))
      done;
      !(fst best ) + 1
    end
      
(* the "cost" of making all categories have rank i under the model *)
let i_cost (i:int) (scores : float array) (b_array : boundaries array) = 
  let res = ref 0. in 
  let f j s =
    let b = b_array.(j) in
    let num_ranks = A.length b + 1 in
      (* normalizing constant *)
    let z = 1. (*b.(A.length b - 1) -. b.(0) *) in
     (*fst boundary case*)
    if i = num_ranks then res := !res +. (max (b.(i-2) -. s) 0.) /. z
      (*snd boundary case*)
    else if i = 1 then res := !res +. (max (s -. b.(0)) 0.) /. z                 
    else let (left,right) = b.(i-2), b.(i-1) in  
	 if s < left then res := !res +. (left -. s) /. z
	 else if s > right then res := !res +. (s -. right) /. z in
  A.iteri f scores ;
  !res

(* returns the second best rank paired with its cost for a given category *)
let second_best (best : int) (score : float) (b : boundaries) : int * float =
  let num_ranks = A.length b + 1 in
    (* normalizing constant *)
  let z = 1. (*b.(num_ranks - 2) -. b.(0)*) in 
  if best = 1 then 
    (2, (b.(0) -. score) /. z)
  else if best = num_ranks then 
    (num_ranks - 1, (score -. b.(num_ranks-2)) /. z)
  else let lower, upper = score -. b.(best-2), b.(best-1) -. score in
       if lower < upper then (best-1, lower /. z) else (best+1, upper /. z)

(* returns true if the distance between ranks is at least i *)
let rnk_dist_i rnks i = 
  let mn, mx = min_max_items rnks in
  mx - mn >= i
         
let dist a1 a2 = 
  let same = all_same a1 in
  if same then 0. 
  else
    Enum.fold2i 
      (fun i n1 n2 accum -> 
	let diff = float (n1 - n2) ** 2. in
(* 	if same then *)
(* 	  accum +. diff *)
(* 	else *)
	accum +. diff +. diff *. float i *. 0.001) 
      0. (A.enum a1) (A.enum a2)

let dfn_rnk_vctrs num_ranks num_categories =
  let rec combine h t = 
    match h, t with
      | _, [] -> []
      | [], _ -> []
      | [a], b :: t' -> A.append a b :: combine h t' 
      | a :: h', t -> (combine [a] t) @ (combine h' t) in
  let rec get n = if n = 1 then [ [|1|] ] else [|n|] :: get (n - 1) in
  let rec get_all c r = 
    let heads = get r in
    if c = 1 then heads
    else 
      let tails = get_all (c - 1) r in
      combine heads tails in
  Global.set all_ranks_glob (A.of_list (get_all num_categories num_ranks))

let get_meta_grf m_scores m_type rnks =
  match m_type with
      Dist i -> begin 
	  let m_score = m_scores.(0) in
	  if (rnk_dist_i rnks i) then 
	    if m_score >= 0. then 0. else -.m_score
	  else
	    if m_score <= 0. then 0. else m_score
	end
    | FoodAtmosDiv ->
	let m_score = m_scores.(0) in
	if rnks.(0) <> rnks.(3) then
	  if m_score >= 0. then 0. else -.m_score
	else
	  if m_score <= 0. then 0. else m_score

let get_rnk_grf rnk score b num_ranks =
  if rnk = num_ranks then
    if score >= b.(rnk-2) then 0. else b.(rnk-2) -. score
  else if rnk = 1 then
    if score <= b.(0) then 0. else score -. b.(0)
  else
    begin
      if score <= b.(rnk-1) then
	if score >= b.(rnk-2) then 0. else b.(rnk-2) -. score
      else score -. b.(rnk-1)
    end
      
let predict_meta m_scores m_type scores b_array =
  let num_ranks, num_categories = 
    A.length b_array.(0) + 1, A.length b_array in
  if not (Global.isdef all_ranks_glob) then
    dfn_rnk_vctrs num_ranks num_categories;
  let all_rnks = Global.get all_ranks_glob in
  let all_grfs = A.map 
    (fun rnks -> 
      let grf = ref (get_meta_grf m_scores m_type rnks) in
      A.iteri
	(fun i rnk -> 
	  let r_grf = get_rnk_grf rnk scores.(i) b_array.(i) num_ranks in
	  grf := !grf +. r_grf)
	rnks;
      !grf)
    all_rnks in
  let i, _ = min_itemi all_grfs in
  A.fast_sort compare all_grfs; 
  if debug && all_grfs.(0) = all_grfs.(1) then print_endline "TIE!!!";
  if debug then
    print_endline (sprintf "%f, %f, ..." all_grfs.(0) all_grfs.(1));
  all_rnks.(i)
  
let predict_agree a_scores defaults scores w_array b_array = 
  let num_ranks, num_categories = 
    A.length b_array.(0) + 1, A.length b_array in 
  if not (Global.isdef all_ranks_glob) then
    dfn_rnk_vctrs num_ranks num_categories;
  let all_rnks = Global.get all_ranks_glob
    (* c_(i,j) := cost of choosing rank j+1 for category i *)
  and cst_mtrx = A.make_matrix num_categories num_ranks 0. in
  A.iteri 
    (fun i def -> 
      for j = def + 1 to num_ranks do
	let cost = b_array.(i).(j-2) -. scores.(i) in
	cst_mtrx.(i).(j-1) <- cost 
      done;
      for j = def - 1 downto 1 do
	let cost = scores.(i) -. b_array.(i).(j-1) in
	cst_mtrx.(i).(j-1) <- cost 
      done;
      cst_mtrx.(i).(def-1) <- 0.)
    defaults;
  let costs = A.make (A.length all_rnks) 0. in 
  A.iteri
    (fun i rnk ->
      (* local costs *)
      A.iteri 
	(fun j r -> costs.(i) <- costs.(i) +. cst_mtrx.(j).(r-1)) 
	rnk;
      (* global costs *)
      L.iter  
	(fun (l, s) ->
	  match l with
	    | hd :: tl -> 
		let same = L.for_all (fun idx -> rnk.(hd) = rnk.(idx)) tl in
		if not same && s > 0. then costs.(i) <- costs.(i) +. s
		else if same && s < 0. then costs.(i) <- costs.(i) -. s
	    | _ -> failwith "huh?")
	a_scores)
    all_rnks;
  let bst_i, bst_cost = ref 0, ref costs.(0) in
  A.iteri
    (fun i c -> 
      if ((c < !bst_cost) || 
	     (c = !bst_cost && dist all_rnks.(i) defaults <= dist all_rnks.(!bst_i) defaults)) 
      then
	(bst_i := i; bst_cost := c))
    costs;
  () ;
  all_rnks.(!bst_i)
    
    
let predict_concensus c_score defaults scores b_array =
  let same, num_categories, num_ranks = 
    all_same defaults, A.length b_array, A.length b_array.(0) in  
  if c_score > 0. && same then 
    (incr aaa; defaults)
  else if c_score > 0. then 
    begin
      let costs = A.init 
	num_ranks 
	(fun i -> i_cost (i+1) scores b_array) in
      let best_i, best_cost = ref 0, ref costs.(0) in
      A.iteri 
	(fun i c -> if c < !best_cost then (best_i := i; best_cost := c))
	costs ;
      if !best_cost < c_score then 
	(incr ada ; A.make num_categories (!best_i + 1))
      else (incr add; defaults)
    end 
  else if not same then (incr ddd; defaults) else 
    begin
      let changes = A.init 
	num_categories 
	(fun i -> second_best defaults.(i) scores.(i) b_array.(i)) in
      let best_i, best_cost, best_new = ref 0, ref infinity, ref 0 in
      A.iteri 
	(fun i (n, cost) -> if cost < !best_cost then 
	    (best_cost := cost; best_i := i; best_new := n))
	changes ;
      if !best_cost < (-. c_score) then 
	(defaults.(!best_i) <- !best_new ; incr dad ; defaults)
      else (incr daa; defaults)
    end

let get_scores_sim instances model = 
  let sim, models = model.sim, model.models in
  let get_score i instance = 
    Enum.foldi
      (fun j m sum ->
	let w = if m.avg && m.trained then m.w_avg else m.w in
	sum +. (dot_prd_fw instance.features w) *. H.find sim (i,j)) 
      0.
      (A.enum models) in
  A.mapi get_score instances
  
    
(* XXX doesn't do voting!!! *)
let predict_sim instances model =
  let scores, models = get_scores_sim instances model, model.models in
  let predict i score = 
    let b = if models.(i).avg && models.(i).trained then models.(i).b_avg else models.(i).b in
    get_rank score b in
  A.mapi predict scores
    
    
let predict_multi ?(a_scores=None) ?(c_score=None) ?(m_scores=None) 
    (instances : instance array) (model : multi_model) =
  if (model.agree = None && c_score = None && a_scores = None && m_scores = None) then 
    if model.use_sim then predict_sim instances model
    else
      A.init (A.length instances) (fun i -> predict instances.(i) model.models.(i))
  else 
    let instance = instances.(0) in
    let features = instances.(0).features 
    and num_categories = model.num_categories
    and num_ranks = model.models.(0).num_ranks in
    if not model.models.(0).vote || not model.multi_trained then
      begin
	let w_array = A.map 
	  (fun m -> if m.avg && m.trained then m.w_avg else m.w) 
	  model.models
	and b_array = A.map
	  (fun m -> if m.avg && m.trained then m.b_avg else m.b) 
	  model.models in
	let scores = if model.use_sim then 
	    get_scores_sim instances model 
	  else
	    A.map 
	      (fun w -> dot_prd_fw features w) 
	      w_array in
	if not (m_scores = None) then
	  predict_meta (Opt.get m_scores) (Opt.get model.m_type) scores b_array
	else
	  let defaults = A.mapi 
	    (fun i s -> get_rank s b_array.(i))
	    scores in 
	  if debug then 
	    (print_newline (); print_string "@@"; 
	     for i=0 to 4 do print_int defaults.(i) done;
	     incr counter; print_newline (); print_int !counter; print_newline ());
	  if a_scores = None then
	    let c_score = match c_score with 
	      | Some s -> s
	      | None -> model.alpha *. Perceptron.get_score instance (Opt.get model.agree) in
	    predict_concensus c_score defaults scores b_array
	  else 
	    let a_scores = (Opt.get a_scores) in 
	    predict_agree a_scores defaults scores w_array b_array
      end
    else (*voting!*) 
      begin
	let w_array = A.map (fun m -> Dray.get m.w_hist 0) model.models 
	and b_array = A.map (fun m -> Dray.get m.b_hist 0) model.models in
	let scores = A.map (fun w -> dot_prd_ff features w) w_array in
	let defaults = A.mapi (fun i s -> get_rank s b_array.(i)) scores 
	and votes = A.make_matrix num_categories num_ranks 0 
	and w_hists = A.map (fun m -> Dray.enum m.w_hist) model.models
	and b_hists = A.map (fun m -> Dray.enum m.b_hist) model.models
	and hist_counts = A.map (fun m -> Dray.enum (deep_copy m.hist_count)) model.models in
	for z = 1 to !(model.num_updates) + 1 do
	  for i = 0 to num_categories - 1 do
	    let count = Opt.get (Enum.peek hist_counts.(i)) in
	    if !count = 0 then 
	      begin
		Enum.junk hist_counts.(i);
		let w,b = Opt.get (Enum.get w_hists.(i)), Opt.get (Enum.get b_hists.(i)) in
		scores.(i) <- scores.(i) +. dot_prd_ff features w;
		defaults.(i) <- get_rank scores.(i) b;
		w_array.(i) <- w;
		b_array.(i) <- b
	      end;
	    let count = Opt.get (Enum.peek hist_counts.(i)) in (*in case it changed*)
	    decr count;
	  done;
	  let predictions = 
	    if a_scores = None then
	      let c_score = match c_score with 
		| Some s -> s
		| None -> model.alpha *. Perceptron.get_score instance (Opt.get model.agree) in
	      predict_concensus c_score defaults scores b_array
	    else 
	      let a_scores = Opt.get a_scores in 
	      predict_agree a_scores defaults scores w_array b_array in
	  A.iteri
	    (fun j p -> votes.(j).(p-1) <- votes.(j).(p-1) + 1)
	    predictions;
	done;
	let get_best j = 
	  let best = (ref 0, ref votes.(j).(0)) in
	  for i = 1 to A.length votes.(j) - 1 do
	    if votes.(j).(i) >= !(snd best) then
	      (fst best := i; snd best := votes.(j).(i))
	  done;
	  !(fst best) + 1 in
	A.init num_categories (fun j -> get_best j) 
      end
	  
				 
let run_model ?(verbose=1) (model' : model) data out_file =
  let model = match model' with
    | Single s_model -> multi_of_single s_model
    | Multi m_model -> m_model in
  let mistakes, distance, category, index = ref 0., ref 0., ref 0, ref 0 in
  let decode ?(c_score=None) ?(a_scores=None) ?(m_scores=None) instances =
    index := !index + A.length instances;
    if !index mod 50 = 0 && verbose > 1 then
      (printf "..%d.." !index; flush stdout ) ;
    let predictions = predict_multi ~c_score ~a_scores ~m_scores instances model in
    A.iteri 
      (fun i n -> fprintf out_file "%d" n ;
	if n <> instances.(i).label then
	  (mistakes := !mistakes +. 1.;
	   distance := 
	     !distance +. float (abs (instances.(i).label - predictions.(i)))))
      predictions;
    output_string out_file "\n" in
  let instances, i = A.make model.num_categories empty_instance, ref 0 in
  let c_scores',a_scores',m_scores' = if model.multi_trained 
    then model.c_test, model.a_test, model.m_test 
    else model.c_train, model.a_train, model.m_train in
  let c_scores = Opt.map_default L.enum (Enum.empty ()) c_scores'
  and a_scores = Opt.map_default L.enum (Enum.empty ()) a_scores' 
  and m_scores = Opt.map_default L.enum (Enum.empty ()) m_scores' in
  L.iter 
    (fun inst -> instances.(!i) <- inst; incr i; 
      if !i = model.num_categories then 
	(i := 0; 
	 decode 
	   ~c_score:(Enum.get c_scores) ~a_scores:(Enum.get a_scores) 
	   ~m_scores:(Enum.get m_scores) instances))
    data;
  let total = float_of_int (L.length data) in
  let err = !mistakes /. total in
  let dist = !distance /. total in
  if verbose > 0 then 
      (printf "test error: %f" err;
       printf "\navg distance: %f\n" dist)
    

      


