(*pp $PP *)

(** Various utility types and functions of my own devising *)
 

module H = ExtHashtbl.Hashtbl
module A = ExtArray.Array
module S = ExtString.String
module L = ExtList.List
module E = Enum
module G = Global
module Opt = Option
module Dray = DynArray

let ( += ) (r : int ref) (n : int) = r := !r + n
let ( -= ) (r : int ref) (n : int) = r := !r - n
let ( *= ) (r : int ref) (n : int) = r := !r * n
let ( /= ) (r : int ref) (n : int) = r := !r / n

let ( +.= ) (r : float ref) (n : float) = r := !r +. n
let ( -.= ) (r : float ref) (n : float) = r := !r -. n
let ( *.= ) (r : float ref) (n : float) = r := !r *. n
let ( /.= ) (r : float ref) (n : float) = r := !r /. n

let ( |> ) x f = f x
let ( >> ) f g x = g(f x)
let ( $ ) f x = f x
let flip f x y = f y x

type label = int
type features = (int,float) H.t
type meta = {global : string; local : string}
type instance = {label : int; features : features; info : meta}
type data = instance list

let empty_instance = {label=0; features=(H.create 0); info={local="";global=""}}

(* XXX now using Std.input_list instead *)
(* let list_of_in in_ch =   *)
(*   let rec f accum =  *)
(*     try *)
(*       f ((input_line in_ch) :: accum) *)
(*     with End_of_file ->  *)
(*       accum in *)
(*   L.rev (f []) *)
    
let list_of_file file_name =
  let in_ch = open_in file_name in
  Std.input_list in_ch

let array_of_input input =
  A.of_enum (Std.input_lines input)
    
let array_of_file file_name = 
  array_of_input (open_in file_name)

let enum_of_file file_name =
  Std.input_lines (open_in file_name)

let output_enum out_ch (enum : string E.t) =
  E.iter (fun s -> output_string out_ch s; output_string out_ch "\n") enum;
  flush out_ch

let file_of_enum file_name (enum : string E.t) =
  let out_ch = open_out file_name in
  output_enum out_ch enum;
  close_out out_ch

let output_list out_ch (list : string list) = 
  output_enum out_ch (L.enum list)
    
let file_of_list file_name (list : string list) = 
  file_of_enum file_name (L.enum list)
  
let output_array out_ch (array : string array) =
  output_enum out_ch (A.enum array)

let file_of_array file_name (array : string array) = 
  file_of_enum file_name (A.enum array)

let entry_add  (tbl : ('a,int) H.t) (key : 'a) (n : int) = 
  if H.mem tbl key then 
    H.replace tbl key ((H.find tbl key) + n)
  else
    H.add tbl key n

let entry_add_ (tbl : ('a,float) H.t) (key : 'a) (n : float) = 
  if H.mem tbl key then 
    H.replace tbl key ((H.find tbl key) +. n)
  else
    H.add tbl key n

let entry_get (tbl : ('a,'b) H.t) (key : 'a) (default : 'b) =
  if H.mem tbl key then 
    H.find tbl key 
  else
    default

let dot_prd_ff (features1 : features) (features2 : features) = 
  let mult feat feat_val sum = 
    if (H.mem features2 feat) then
      feat_val *. (H.find features2 feat) +. sum
    else sum in
  H.fold mult features1 0.

let dot_prd_fw (features : features) (weights : float array) = 
  let mult feat feat_val sum = 
    try
      weights.(feat) *. feat_val +. sum
    with Invalid_argument s -> sum in
  H.fold mult features 0.

let split (s : string) = Str.split (Str.regexp "[ \t\n\r]+") s

let min_itemi a =
  let l = A.length a in
  if l = 0 then failwith "array of length 0"
  else
    let min_i, min_e = ref 0, ref a.(0) in
    for i = 1 to l - 1 do
      let e = a.(i) in
      if e < !min_e then 
  (min_i := i; min_e := e)
    done;
    !min_i, !min_e

let min_item a = snd (min_itemi a)
  
let max_itemi a =
  let l = A.length a in
  if l = 0 then failwith "array of length 0"
  else
    let max_i, max_e = ref 0, ref a.(0) in
    for i = 1 to l - 1 do
      let e = a.(i) in
      if e > !max_e then 
  (max_i := i; max_e := e)
    done;
    !max_i, !max_e
  
let max_item a = snd (max_itemi a)

let min_max_items a = 
  let l = A.length a in
  if l = 0 then failwith "array of length 0"
  else
    let mx, mn = ref a.(0), ref a.(0) in
    for i = 1 to l - 1 do
      let e = a.(i) in
      if e > !mx then mx := e;
      if e < !mn then mn := e
    done;
    !mn, !mx

(** [unique a] returns a new array [a'] with all the elements of [a] appearing
    exactly once, in the intial order of appearance in [a]. *)
let unique a = 
  let h = H.create (A.length a * 2) 
  and n = ref 0 in
  A.iteri 
    (fun i e -> 
      if H.mem h e then n += 1
      else H.add h e (i - !n)) 
    a;
  let a' = A.make (H.length h) in
  H.iter (fun e i -> a.(i) <- e) h;
  a'
 
let sum a1 a2 = 
  let l1, l2 = A.length a1, A.length a2 in
  if l1 <> l2 then failwith "arrays of unequal length"
  else
    A.init l1 (fun i -> a1.(i) +. a2.(i))

let round x = int_of_float (floor (x +. 0.5))

let parse_features (fs_list : string list) =
  let tbl = H.create 15 in
  let add (fs : string) =
    match Str.split (Str.regexp ":") fs with
      | [f; f_val] ->   
	  (try H.replace tbl (int_of_string f) (float_of_string f_val)
	   with _ -> failwith ("couldn't parse(1): " ^ fs))
      | _ -> failwith "Wrong feature format!!" in
  L.iter add fs_list; 
  tbl
 
(* XXX in multi-mode, assumes each label is only one character *)
let parse_data_line ?(global="") ?(m=1) (s : string) : instance list =
  let s,info = 
    match Str.bounded_split (Str.regexp "#") s 2 with
      | [fs; cs] -> S.strip fs, {global = global; local = S.strip cs}
      | [fs] -> S.strip fs, {global = global; local = ""} 
      | _ -> failwith "encountered blank line!" in
  if m = 1 then  
    match split s with
      | label :: fs_list -> 
	  (try [{ label = int_of_string label; 
		  features = parse_features fs_list;
		  info = info }] 
	   with _ -> failwith ("couldn't parse(2): " ^ s))
      | _ -> failwith ("wrong input format!! " ^ s)
  else 
    match split s with
      | labels :: fs_list ->
	  let features = parse_features fs_list in
	  let rec get_instances s' = 
	    let s, neg = 
              if S.starts_with s' "-" then 
		(String.sub s' 1 (String.length s' - 1), -1)
              else
		(s', 1) in
	    if String.length s = 1 then 
              (try [{ label = neg * int_of_string s; 
		      features = features;
		      info = info }]
	       with _ -> failwith ("couldn't parse(3): " ^ s))
	    else
              (try
		 { label = neg * int_of_string (String.sub s 0 1); 
		   features = features;
		   info = info } 
		 :: get_instances (Str.string_after s 1) 
	       with _ -> failwith ("couldn't parse(4): " ^ s)) in
	  let res = get_instances labels in 
	  if L.length res = m then
	    res
	  else
	    failwith ("didn't find " ^ string_of_int m ^ " single-character labels!!" ^ s)
      | _ -> failwith ("wrong input format(2)!!" ^ s)
	  
let parse_data_file ?(verbose=1) ?(multi=false) data_file : data * int = 
  let global = ref "" in
  let rec get_next_line df = 
    let res = S.strip (input_line df) in
    if res.[0] = '.' then 
      begin
	(match Str.bounded_split (Str.regexp ".") res 1 with
	   | [g] -> global := S.strip g
	   | [] -> global := ""
	   | _ -> failwith "shouldn't happen 33");
	get_next_line df
      end
    else res in
  let fst_ln = get_next_line data_file in
  let m = if multi then  
    String.length (Str.global_replace (Str.regexp "-") "" (L.hd (split fst_ln)))  
  else 1 in
  let instances = Dray.make 1000 in
  L.iter (Dray.add instances) (parse_data_line ~global:!global ~m fst_ln);
  let count = ref 0 in
  (try while true do
     (if !count mod 50 = 0 && verbose > 1then
	(print_int !count; print_string "...");
	incr count;
	let ln = get_next_line data_file in
	L.iter (Dray.add instances) (parse_data_line ~global:!global ~m ln)) 
   done
   with End_of_file -> ());
  flush stdout;
  (Dray.to_list instances), m
    (* let rec go (accum : instance list) =  *)
(*     try *)
(*       let ln = get_next_line data_file in *)
(*       go ((parse_data_line ~global:!global ~m ln) @ accum) *)
(*     with End_of_file -> accum in *)
(*   (L.rev (go (parse_data_line ~global:!global ~m fst_ln)), m) *)

let get_labels (data : data) : label array =
  let l = H.create 15 in
  let f inst = H.replace l inst.label 1 in
  L.iter f data;
  A.of_enum (H.keys l)

let get_highest_feature (data : data) : int = 
  let highest_in_instance (instance : instance) =
    let f feat feat_val prev_high = max feat prev_high in
    H.fold f instance.features 0 in
  L.fold_left max 0 (L.map highest_in_instance data)

let sign a = if a < 0 then 1 else -1

let get_mean_dev l =
  let sum, num = L.fold_left ( +. ) 0. l, float_of_int (L.length l) in
  let mean = sum /. num in
  let dev = sqrt ((L.fold_left (fun sum v -> sum +. (v -. mean)**2.) 0. l) /. num) in
  (mean, dev)
    
let all_same a = 
  let rec samei a2 i = 
    if i = A.length a2 then true else a2.(0) = a2.(i) && samei a2 (i + 1) in
  samei a 1

let id x = x

let deep_copy da = 
  let l = Dray.length da in
  Dray.init l (fun i -> ref !(Dray.get da i)) 

let shuffle_array ?(seed=None) a =
  if A.length a = 0 then 
    raise (Invalid_argument "can't shuffle an empty array");
  (match seed with 
    | None -> Random.self_init ()
    | Some n -> Random.init n);
  let rec fy a =  function  (* fisher-yates shuffle algorithm *)
    | 0 -> a
    | i -> let swap_pos = Random.int i in
      let temp = a.(i) in
      a.(i) <- a.(swap_pos);
      a.(swap_pos) <- temp;
      fy a (pred i) in
  fy a (A.length a - 1)

let shuffle_list ?(seed=None) l =
  A.to_list (shuffle_array (A.of_list l))

 (** 
     Takes a multi hashtable h and returns a new Hashtable with exactly one entry
     for each key in h, mapped to an array of all bindings of that key in h. 
 *)
let mht_to_ht (h : ('a, 'b) Hashtbl.t) : ('a, 'b array) Hashtbl.t =
  let h' = H.create (H.length h) in
  Enum.iter 
    (fun k -> H.replace h' k (A.of_list (H.find_all h k)))
    (H.keys h);
  h'


  
    
  
